@@ -6,6 +6,8 @@
 
 `getQuote` <-
 function(Symbols,src='yahoo',what, ...) {
+  importDefaults("getQuote")
+
   Symbols <- unique(unlist(strsplit(Symbols,";")))
   args <- list(Symbols=Symbols,...)
   if(!missing(what))
@@ -22,9 +24,54 @@ function(Symbols,src='yahoo',what, ...) {
   df[Symbols,]
 }
 
+.yahooSession <- function(is.retry = FALSE) {
+  cache.name <- "_yahoo_curl_session_"
+  ses <- get0(cache.name, .quantmodEnv) # get cached session
+  
+  if (is.null(ses) || is.retry) {
+    ses <- list()
+    ses$h <- curl::new_handle()
+    # yahoo finance doesn't seem to set cookies without these headers
+    # and the cookies are needed to get the crumb
+    curl::handle_setheaders(ses$h, accept = "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7")
+    URL <- "https://finance.yahoo.com/"
+    r <- curl::curl_fetch_memory(URL, handle = ses$h)
+    # yahoo redirects to a consent form w/ a single cookie for GDPR:
+    # detecting the redirect seems very brittle as its sensitive to the trailing "/"
+    ses$can.crumb <- ((r$status_code == 200) && (URL == r$url) && (NROW(curl::handle_cookies(ses$h)) > 1))
+    assign(cache.name, ses, .quantmodEnv) # cache session
+  }
+
+  if (ses$can.crumb) {
+    # get a crumb so that downstream callers don't have to handle invalid sessions.
+    # this is a network hop, but very lightweight payload
+    n <- if (unclass(Sys.time()) %% 1L >= 0.5) 1L else 2L
+    query.srv <- paste0("https://query", n, ".finance.yahoo.com/v1/test/getcrumb")
+    r <- curl::curl_fetch_memory(query.srv, handle = ses$h)
+    if ((r$status_code == 200) && (length(r$content) > 0)) {
+      ses$crumb <- rawToChar(r$content)
+    } else {
+      # we were unable to get a crumb
+      if (is.retry) {
+        # we couldn't get a crumb with a new session
+        stop("unable to get yahoo crumb")
+      } else {
+        # we tried to re-use a session but couldn't get a crumb
+        # try to get a crumb using a new session
+        ses <- .yahooSession(TRUE)
+      }
+    }
+  }
+
+  return(ses)
+}
+
 `getQuote.yahoo` <-
-function(Symbols,what=standardQuote(),...) {
+function(Symbols,what=standardQuote(),session=NULL,...) {
+  importDefaults("getQuote.yahoo")
   length.of.symbols <- length(Symbols)
+  if (is.null(session)) session <- .yahooSession()
+  
   if(length.of.symbols > 200) {
     # yahoo only works with 200 symbols or less per call
     # we will recursively call getQuote.yahoo to handle each block of 200
@@ -35,12 +82,14 @@ function(Symbols,what=standardQuote(),...) {
     for(i in 1:length(all.symbols)) {
       Sys.sleep(0.5)
       cat(i,", ")
-      df <- rbind(df, getQuote.yahoo(all.symbols[[i]],what))
+      df <- rbind(df, getQuote.yahoo(all.symbols[[i]],what,session=session))
     }
     cat("...done\n")
     return(df)
   }
-  SymbolsString <- paste(Symbols,collapse=',')
+  # escape symbols that have special characters
+  escapedSymbols <- sapply(Symbols, URLencode, reserved = TRUE)
+  SymbolsString <- paste(escapedSymbols, collapse = ',')
   if(inherits(what, 'quoteFormat')) {
     QF <- what[[1]]
     QF.names <- what[[2]]
@@ -53,11 +102,15 @@ function(Symbols,what=standardQuote(),...) {
   # exchange, fullExchangeName, market, sourceInterval, exchangeTimezoneName,
   # exchangeTimezoneShortName, gmtOffSetMilliseconds, tradeable, symbol
   QFc <- paste0(QF,collapse=',')
-  URL <- paste0("https://query1.finance.yahoo.com/v7/finance/quote?symbols=",
-                SymbolsString,
-                "&fields=",QFc)
+  if (session$can.crumb) {
+    URL <- paste0("https://query1.finance.yahoo.com/v7/finance/quote?crumb=", session$crumb)
+  } else {
+    URL <- "https://query1.finance.yahoo.com/v6/finance/quote?"
+  }
+
+  URL <- paste0(URL, "&symbols=", SymbolsString, "&fields=", QFc)
   # The 'response' data.frame has fields in columns and symbols in rows
-  response <- jsonlite::fromJSON(curl::curl(URL))
+  response <- jsonlite::fromJSON(curl::curl(URL, handle = session$h))
   if (is.null(response$quoteResponse$error)) {
     sq <- response$quoteResponse$result
   } else {
@@ -300,29 +353,37 @@ getQuote.av <- function(Symbols, api.key, ...) {
 
   # Function to process each quote response
   quote2df <-
-    function(response, map)
+    function(response, map, symbol)
   {
     # Expected response structure
     qres <- setNames(vector("list", nrow(map)), map[["av.names"]])
 
-    elem <- function(el)
+    elem <- function(el, isnum)
     {
-      # process numeric columns
       res <- NA_real_
       if (!is.null(el)) {
-        haspct <- grepl("%", el, fixed = TRUE)
-        if (haspct) {
-          el <- sub("%", "", el, fixed = TRUE)
-          res <- as.numeric(el) / 100
+        if (isnum) {
+          # process numeric columns
+          haspct <- grepl("%", el, fixed = TRUE)
+          if (haspct) {
+            el <- sub("%", "", el, fixed = TRUE)
+            res <- as.numeric(el) / 100
+          } else {
+            res <- as.numeric(el)
+          }
         } else {
-          res <- as.numeric(el)
+          res <- el
         }
       }
       res
     }
     tmp <- modifyList(qres, response)
-    tonum <- map[["is.number"]]
-    tmp[tonum] <- lapply(tmp[tonum], elem)
+    tmp <- Map(elem, el = tmp, isnum = map[["is.number"]])
+
+    # populate Symbol column for symbols missing quotes
+    if (is.na(tmp[["01. symbol"]])) {
+      tmp[["01. symbol"]] <- symbol
+    }
 
     data.frame(tmp, stringsAsFactors = FALSE)
   }
@@ -344,7 +405,7 @@ getQuote.av <- function(Symbols, api.key, ...) {
               call. = FALSE, immediate. = TRUE)
     } else {
       resp <- resp[[1]]  # resp$`Global Quote`
-      qlist[[Symbol]] <- quote2df(resp, map)
+      qlist[[Symbol]] <- quote2df(resp, map, Symbol)
     }
   }
 
